(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('jsen')) :
  typeof define === 'function' && define.amd ? define(['jsen'], factory) :
  (global.JsonSchemaValidationStrategy = factory(global.jsen));
}(this, (function (jsen) { 'use strict';

jsen = 'default' in jsen ? jsen['default'] : jsen;

function followPath(schema, step) {
  if (schema.type === 'object') {
    return schema.properties[step];
  } else if (schema.type === 'array') {
    return followPath(schema.items || {}, step);
  }
}

function inspectSchema (schema, key) {
  var path = key.replace(/\.[0-9]+\./, '.').replace(/\.[0-9]+$/, '').split('.');
  var context = schema;
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = path[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var step = _step.value;

      context = followPath(context, step);
      if (!context) {
        break;
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return context;
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();

var defaultMessages = {
  'type': '"{{key}}" must be a {{schema.type}}',
  'enum': '"{{key}}" must be one of: {{schema.enum}}',
  'minimum': '"{{key}}" must be less than or equal to {{schema.minimum}}',
  'exclusiveMinimum': '"{{key}}" must be less than {{schema.exclusiveMinimum}}',
  'maximum': '"{{key}}" must be larger than or equal to {{schema.maximum}}',
  'exclusiveMaximum': '"{{key}}" must be larger than or equal to {{schema.exclusiveMaximum}}',
  'multipleOf': '"{{key}}" must be multiple of {{schema.multipleOf}}',
  'minLength': '"{{key}}" length must be at least {{schema.minLength}} characters long',
  'maxLength': '"{{key}}" length must be less than or equal to {{schema.maxLength}} characters long',
  'pattern': '"{{key}}" should match pattern "{{schema.pattern}}"',
  'format': '"{{key}}" must be a valid {{schema.format}}',
  'minItems': '"{{key}}" must contain at least {{schema.minItems}} items',
  'maxItems': '"{{key}}" must contain less than or equal to {{schema.maxItems}} items',
  'additionalItems': '"{{key}}" is invalid',
  'uniqueItems': '"{{key}}" must contain unique items',
  'items': '"{{key}}" is invalid',
  'minProperties': '"{{key}}" must contain at least {{schema.minProperties}} properties',
  'maxProperties': '"{{key}}" must contain less than or equal to {{schema.maxProperties}} items',
  'required': '"{{key}}" is required',
  'properties': '"{{key}}" is invalid',
  'patternProperties': '"{{key}}" properties should match pattern "{{schema.patternProperties}}"',
  'additionalProperties': '"{{key}}" is invalid',
  'dependencies': '"{{key}}" is invalid',
  'allOf': '"{{key}}" is invalid',
  'anyOf': '"{{key}}" is invalid',
  'oneOf': '"{{key}}" is invalid',
  'not': '"{{key}}" is invalid',
  '$ref': '"{{key}}" is invalid'
};

function compileMessage(error, template) {
  var prefix = arguments.length <= 2 || arguments[2] === undefined ? '' : arguments[2];

  if (typeof template === 'function') {
    return template(error);
  }
  var result = template;
  var item = void 0;
  for (var key in error) {
    item = error[key];
    if (key === 'enum' && Array.isArray(item)) {
      item = item.join(', ');
    }
    if ((typeof item === 'undefined' ? 'undefined' : _typeof(item)) === 'object') {
      result = compileMessage(error[key], result, key);
    } else {
      result = result.replace(new RegExp('{{' + (prefix ? prefix + '.' + key : key) + '}}', 'g'), item);
    }
  }
  return result;
}

function buildMessage(rawError, messages, schema) {
  var propertySchema = inspectSchema(schema, rawError.path);
  if (propertySchema.type === 'array' && /\.[0-9]+$/.test(rawError.path)) {
    propertySchema = propertySchema.items;
  }
  var error = Object.assign(rawError, {
    schema: propertySchema,
    key: propertySchema && propertySchema.label ? propertySchema.label : rawError.path.split('.').pop()
  });
  var template = messages[error.keyword];
  if (error.message) {
    template = error.message;
  }
  return compileMessage(error, template);
}

function strategy(mainOptions) {
  mainOptions = mainOptions || {};
  mainOptions = Object.assign({
    greedy: true,
    additionalProperties: true,
    messages: Object.assign({}, defaultMessages, mainOptions.messages || {})
  }, mainOptions);

  var jsenOptions = {
    greedy: mainOptions.greedy,
    additionalProperties: mainOptions.additionalProperties
  };

  var makeGetterKey = function makeGetterKey(key) {
    return key.replace(/^\./, '').replace(/\[([0-9]+)\]/, '.$1');
  };

  var makePath = function makePath(key) {
    return makeGetterKey(key).split('.');
  };

  var pathFinder = function pathFinder(path, data) {
    var actual = data;
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = makeGetterKey(path).split('.')[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var key = _step.value;

        actual = actual[key];
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    return actual;
  };

  var schemaPathFinder = function schemaPathFinder(schema, key, data) {
    var lastKey = makePath(key).pop();
    var value = pathFinder(key, data);
    var result = value;
    if (schema.type === 'array' && !isNaN(lastKey)) {
      result = [value];
    }
    return result;
  };

  var getSchema = function getSchema(schema) {
    var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
    var key = options.key;

    if (!key) {
      return [jsen(schema, jsenOptions), function (data) {
        return data;
      }, schema];
    } else {
      var _ret = function () {
        var schemaKey = key.replace(/\[[0-9]+\]/g, '');
        var path = schemaKey.split('.');
        var pathSize = path.length;
        var parentPath = path.slice(0, pathSize - 1);
        var parentKey = parentPath.join('.');
        var propertyName = path[pathSize - 1];
        var parentSchema = parentKey ? inspectSchema(schema, parentKey) : schema;
        var required = parentSchema.required && parentSchema.required.indexOf(propertyName) >= 0 ? [propertyName] : [];
        var fieldSchema = inspectSchema(schema, schemaKey);
        if (!fieldSchema) {
          return {
            v: null
          };
        }
        var partialSchema = {
          type: 'object',
          properties: defineProperty({}, propertyName, fieldSchema)
        };
        if (required.length) {
          partialSchema.required = required;
        }
        return {
          v: [jsen(partialSchema, jsenOptions), function (data) {
            return defineProperty({}, propertyName, schemaPathFinder(fieldSchema, key, data));
          }, partialSchema]
        };
      }();

      if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
    }
  };

  var parseError = function parseError(error, parentKey, schema) {
    var key = error.path;
    var innerPath = key.replace(/^\./, '').replace(/\[([0-9]+)\]/, '.$1').split('.');
    var parentPath = [];
    var parentPathLastKey = void 0;
    if (parentKey) {
      parentPath = makePath(parentKey);
      parentPathLastKey = parentPath.pop(); // intentionally remove
      if (!isNaN(parentPathLastKey)) {
        innerPath[1] = parentPathLastKey;
        parentPath.pop();
      }
    }
    return {
      path: key ? parentPath.concat(innerPath) : parentPath,
      message: buildMessage(error, mainOptions.messages, schema)
    };
  };

  var ensureParentKey = function ensureParentKey(obj, parentKey) {
    if (!parentKey) return obj;
    var context = obj;
    var path = makePath(parentKey);
    var key = void 0;
    for (var i = 0, c = path.length; i < c; i++) {
      key = path[i];
      if (!context[key]) {
        context[key] = i < c - 1 ? {} : undefined;
        context = context[key];
      }
    }
    return obj;
  };

  var parseErrors = function parseErrors(errors, parentKey, schema) {
    var results = ensureParentKey({}, parentKey);
    var result = void 0,
        step = void 0,
        context = void 0;
    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
      for (var _iterator2 = errors[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
        var error = _step2.value;

        result = parseError(error, parentKey, schema);
        context = results;
        for (var i = 0, c = result.path.length; i < c; i++) {
          step = result.path[i];
          if (!results[step]) {
            context[step] = i < c - 1 ? {} : [];
          }
          if (i === c - 1) {
            context[step].push(result.message);
          }
          context = context[step];
        }
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return) {
          _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }

    return results;
  };

  return {
    validate: function validate() {
      var data = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
      var rootSchema = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
      var options = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];
      var callback = arguments[3];
      var key = options.key;

      var results = getSchema(rootSchema, options);
      var errors = ensureParentKey({}, key);
      if (!results) {
        callback(errors);
        return;
      }

      var _results = slicedToArray(results, 3);

      var validate = _results[0];
      var transformer = _results[1];
      var schema = _results[2];

      var valid = validate(transformer(data));
      if (valid) {
        callback(errors);
      } else {
        callback(parseErrors(validate.errors, key, schema));
      }
    }
  };
}

return strategy;

})));
//# sourceMappingURL=json-schema-validation-strategy.js.map
